;; 
;; fix for #'markov-analyze to print to *om-stream*
;;
;; Time-stamp: <2021-02-23 11:08:29 andersvi>
;; 

(defun markov-analyze (seq &key (order 1)
			     (print? t)			    ; t pattern table
			     (pattern? t)		    ; nil or pattern
			     sort?
			     (print-decimals 3) (period nil) key
			     returning)
  (let ((len (length seq))
        (labels '())					    ; the set of all outcomes 
        (table '())
        (row-label-width 8)
        (pat nil)
        (field (+ print-decimals 2)))			    ; n.nnn 
    (labels ((add-outcome (prev next)
	       (let ((entry
		       (find-if #'(lambda (x) (equal prev (car x)))
				table)))
                 (if (not entry)
		     (push (list prev
                                 (format nil "~s" prev)
                                 (list next 1))
                           table)
		     (let ((e (assoc next (cddr entry))))
		       (if e
			   (rplaca (cdr e) (+ 1 (cadr e)))
			   (rplacd (last (cdr entry))
				   (list (list next 1))))))))
	     (before? (x y l)
	       (if (null x)
                   t
		   (let ((p1
			   (position-if #'(lambda (z) (equal (car x) z))
					l))
			 (p2
			   (position-if #'(lambda (z) (equal (car y) z))
					l)))
		     (cond ((< p1 p2) t)
			   ((= p1 p2) (before? (cdr x) (cdr y) l))
			   (t nil)))))
	     (liststring (l)
	       (if (null l)
                   ""
		   (let ((a (format nil "~s" (car l))))
		     (do ((x (cdr l) (cdr x)))
			 ((null x) a)
		       (setf a
			     (concatenate 'string
					  a
					  (format nil
						  " ~s"
						  (car x)))))))))
      (dotimes (i len)
        (loop
	  with prev = (list)
	  for j to order
	  for x = (let ((raw (elt seq (mod (+ i j) len))))
		    (if key (funcall key raw) raw))
	  when (< j order) do (push x prev)
	    finally (progn
		      (add-outcome (reverse prev) x)
		      (if (not (member x labels))
			  (push x labels)))))
      
      ;; sort the outcomes according to user specification: a list, a sorting
      ;; function or nil.
      
      (cond ((consp sort?) (setf labels sort?))
	    (sort? (setf labels (sort labels sort?)))
	    ((numberp (car labels)) (setf labels (sort labels #'<)))
	    ((and (car labels)
                  (let ((t4 (car labels))) (and t4 (symbolp t4))))
	     (setf labels
                   (sort labels
                         #'(lambda (x y)
			     (string-lessp (format nil "~a" x)
					   (format nil "~a" y))))))
	    (t (setf labels (reverse labels))))
      ;; map over data, normalize weights 
      (loop
	for row in table
	for lab = (cadr row)				    ; label
	for val = (cddr row)
	maximize (length lab) into len
	do (let ((total (loop for e in val sum (cadr e))))
	     (setf total (* total 1.0))
	     (loop
	       for e in val
	       do (rplaca (cdr e) (decimals (/ (cadr e) total) print-decimals))))
	finally (setf row-label-width (max len row-label-width)))

      ;; sort table according to value order
      (setf table
	    (sort table
                  #'(lambda (x y) (before? (car x) (car y) labels))))
      (let ((*standard-output* om::*om-stream*))
	(when (member print? '(t table :table))
          (let* ((sp " ")
		 (ln (make-string field :initial-element #\-)))
	    ;; print column header row
	    (progn
	      (terpri)
	      ;; print left pad for row label
	      (dotimes (i row-label-width) (write-char #\*))
	      (dolist (l labels)
		(princ sp)
		(let* ((s (format nil "~a" l)) (n (length s)))
                  (dotimes (i (max (- field n) 0))
		    (write-char #\Space))
                  (princ s))))
	    (dolist (row table)
	      (terpri)
	      (let* ((s (liststring (car row))) (n (length s)))
		;; pad number
		(dotimes (i (max (- row-label-width n) 0))
                  (write-char #\Space))
		(dotimes (i (min row-label-width n))
                  (write-char (elt s i))))
	      (dolist (l labels)
		(let ((v (assoc l (cddr row))))
                  (if (not v)
		      (progn (princ sp) (princ ln))
		      (let* ((s (prin1-to-string (cadr v)))
			     (n (length s)))
			(princ sp)
			(dotimes (i (max (- field n) 0))
			  (write-char #\Space))
			(princ s))))))
	    (terpri))
          (force-output))
	(when (or pattern? (member print? '(t pattern :pattern)))
          (setf pat
		(loop
		  for row in table
		  collect (append (car row) '(->) (cddr row)))))
	(when (member print? '(t pattern :pattern))
          (pprint `(new markov of ',pat))
          (terpri)
	  (force-output)))
      (when pattern?
	;; patterns not defined yet, cant use new or <markov>
	(make-instance (find-class 'markov) :of pat :for period :returning returning)))))
